#include <stdio.h>
#include <stdarg.h>
#include <FL/Fl.H>
#include <FL/x.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Text_Buffer.H>
#ifdef __APPLE__
#  include <OpenGL/gl3.h>
#  include <OpenGL/OpenGL.h>
#elif defined(WIN32)
#  if defined(__MINGW32__) || defined(__CYGWIN__)
#    define GLEW_STATIC 1
#  else
#    define GLEW_BUILD 1
#  endif
#  include <GL/glew.h>
#else
#  include <GL/glew.h>
#endif


void add_output(const char *format, ...);


static void glew_initialize() {
  static bool first = true;
  if (first) {
      first = false;
#ifndef __APPLE__
      GLenum err = glewInit(); // defines pters to functions of OpenGL V 1.2 and above
      add_output("glewInit=%d\n",err);
#endif
      add_output("GL_VERSION=%s\n",(const char *)glGetString(GL_VERSION));
  }
}

class SimpleGL3Window : public Fl_Gl_Window {
  GLuint shaderProgram;
  GLuint vertexArrayObject;
  GLuint vertexBuffer;
  
  GLint positionUniform;
  GLint colourAttribute;
  GLint positionAttribute;
  int gl_version_major;
public:
  SimpleGL3Window(int w, int h) :  Fl_Gl_Window(w, h) {
#ifdef __APPLE__
    static int array[] = { // equivalent to FL_RGB8 | FL_DOUBLE
      kCGLPFADoubleBuffer,
      kCGLPFAOpenGLProfile,
      kCGLOGLPVersion_3_2_Core,
      //kCGLOGLPVersion_GL4_Core,
      0
    };
    mode(array);
#else
    mode(FL_RGB8 | FL_DOUBLE);
#endif
    shaderProgram = 0;
  }
  void draw() {
    if (!shaderProgram) {
      glew_initialize();
      sscanf((const char *)glGetString(GL_VERSION), "%d", &gl_version_major);
      if (gl_version_major < 2) return;
      GLuint  vs;
      GLuint  fs;
      int Mslv, mslv; // major and minor version numbers of the shading language
      sscanf((char*)glGetString(GL_SHADING_LANGUAGE_VERSION), "%d.%d", &Mslv, &mslv);
      add_output("Shading Language Version=%d.%d\n",Mslv, mslv);
      const char *vss_format="#version %d%d\n\
      uniform vec2 p;\
      in vec4 position;\
      in vec4 colour;\
      out vec4 colourV;\
      void main (void)\
      {\
      colourV = colour;\
      gl_Position = vec4(p, 0.0, 0.0) + position;\
      }";
      char vss_string[300]; const char *vss = vss_string;
      sprintf(vss_string, vss_format, Mslv, mslv);
      const char *fss_format="#version %d%d\n\
      in vec4 colourV;\
      out vec4 fragColour;\
      void main(void)\
      {\
      fragColour = colourV;\
      }";
      char fss_string[200]; const char *fss = fss_string;
      sprintf(fss_string, fss_format, Mslv, mslv);
      GLint err; GLchar CLOG[1000]; GLsizei length;
      vs = glCreateShader(GL_VERTEX_SHADER);
      glShaderSource(vs, 1, &vss, NULL);
      glCompileShader(vs);
      glGetShaderiv(vs, GL_COMPILE_STATUS, &err);
      if (err != GL_TRUE) {
	glGetShaderInfoLog(vs, sizeof(CLOG), &length, CLOG); 
	add_output("vs ShaderInfoLog=%s\n",CLOG);
	}
      fs = glCreateShader(GL_FRAGMENT_SHADER);
      glShaderSource(fs, 1, &fss, NULL);
      glCompileShader(fs);
      glGetShaderiv(fs, GL_COMPILE_STATUS, &err);
      if (err != GL_TRUE) {
	glGetShaderInfoLog(fs, sizeof(CLOG), &length, CLOG); 
	add_output("fs ShaderInfoLog=%s\n",CLOG);
	}     
      // Attach the shaders
      shaderProgram = glCreateProgram();
      glAttachShader(shaderProgram, vs);
      glAttachShader(shaderProgram, fs);
      if (gl_version_major >=3) glBindFragDataLocation(shaderProgram, 0, "fragColour");
      glLinkProgram(shaderProgram);
      glGetProgramiv(shaderProgram, GL_LINK_STATUS, &err);
      if (err != GL_TRUE) {
        glGetProgramInfoLog(shaderProgram, sizeof(CLOG), &length, CLOG);
        add_output("link log=%s\n", CLOG);
      }
      // Get pointers to uniforms and attributes
      positionUniform = glGetUniformLocation(shaderProgram, "p");
      colourAttribute = glGetAttribLocation(shaderProgram, "colour");
      positionAttribute = glGetAttribLocation(shaderProgram, "position");
      glDeleteShader(vs);
      glDeleteShader(fs);
      add_output("positionUniform: %i, colourAttribute: %i, positionAttribute: %i\n",positionUniform,colourAttribute,positionAttribute);
      
      // Upload vertices (1st four values in a row) and colours (following four values)
      GLfloat vertexData[]= { -0.5,-0.5,0.0,1.0,   1.0,0.0,0.0,1.0,
        -0.5, 0.5,0.0,1.0,   0.0,1.0,0.0,1.0,
        0.5, 0.5,0.0,1.0,   0.0,0.0,1.0,1.0,
        0.5,-0.5,0.0,1.0,   1.0,1.0,1.0,1.0};
      if (gl_version_major >=3) glGenVertexArrays(1, &vertexArrayObject);
      if (gl_version_major >=3) glBindVertexArray(vertexArrayObject);
      
      glGenBuffers(1, &vertexBuffer);
      glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
      glBufferData(GL_ARRAY_BUFFER, 4*8*sizeof(GLfloat), vertexData, GL_STATIC_DRAW);
      
      glEnableVertexAttribArray((GLuint)positionAttribute);
      glEnableVertexAttribArray((GLuint)colourAttribute  );
      glVertexAttribPointer((GLuint)positionAttribute, 4, GL_FLOAT, GL_FALSE, 8*sizeof(GLfloat), 0);
      glVertexAttribPointer((GLuint)colourAttribute  , 4, GL_FLOAT, GL_FALSE, 8*sizeof(GLfloat), (char*)0+4*sizeof(GLfloat));
    }
    else if ((!valid())) {
#if FL_MAJOR_VERSION*1000+FL_MINOR_VERSION*100+FL_PATCH_VERSION > 10304
      glViewport(0, 0, pixel_w(), pixel_h());
#else
      glViewport(0, 0, w(), h());
#endif
    }
    glClearColor(0.08, 0.8, 0.8, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);
    glUseProgram(shaderProgram);
    GLfloat p[]={0,0};
    glUniform2fv(positionUniform, 1, (const GLfloat *)&p);
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
  }
  int handle(int event) {
    if (event == FL_PUSH && gl_version_major >= 2) {
      static float factor = 1.1;
      GLfloat data[4];
      glGetBufferSubData(GL_ARRAY_BUFFER, 0, 4*sizeof(GLfloat), data);
      if (data[0] < -0.88 || data[0] > -0.5) factor = 1/factor;
      data[0] *= factor;
      glBufferSubData(GL_ARRAY_BUFFER, 0, 4*sizeof(GLfloat), data);
      glGetBufferSubData(GL_ARRAY_BUFFER, 24*sizeof(GLfloat), 4*sizeof(GLfloat), data);
      data[0] *= factor;
      glBufferSubData(GL_ARRAY_BUFFER, 24*sizeof(GLfloat), 4*sizeof(GLfloat), data);
      redraw();
      add_output("push\n");
      return 1;
    }
    return Fl_Gl_Window::handle(event);
  }
};

Fl_Text_Display *output; // shared between output_win() and add_output()

void output_win()
{
  Fl_Window *win = new Fl_Window(500,500,"output");
  output = new Fl_Text_Display(0,0,win->w(), win->h());
  win->end();
  output->buffer(new Fl_Text_Buffer());
  win->show();
}


void add_output(const char *format, ...)
{
  va_list args;
  char line_buffer[10000];
  va_start(args, format);
  vsnprintf(line_buffer, sizeof(line_buffer)-1, format, args);
  va_end(args);
  output->insert(line_buffer);
}


int main(int argc, char **argv)
{
#if defined(__APPLE__) && FL_MAJOR_VERSION*1000+FL_MINOR_VERSION*100+FL_PATCH_VERSION > 10304
  Fl::use_high_res_GL(1);
#endif
  SimpleGL3Window *win = new SimpleGL3Window(500, 500);
  win->end();
  win->resizable(win);
  win->label("Click to reshape");
  output_win();
  win->show();
  Fl::run();
}
